#include <lektor/common.h>
#include <lektor/queue.h>
#include <lektor/commands.h>
#include <lektor/database.h>
#include <lektor/net.h>
#include <lektor/reg.h>

#include <signal.h>
#include <sched.h>
#include <netdb.h>
#include <poll.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <sys/socket.h>

/* Server commands, stored in a trie for a bit more efficient lookups.
 * Commands have ascii names, but use the size of a char to not bother with
 * bounds. If the `type` is not LKT_COMMAND_NULL, then the node is a terminal
 * node, otherwise you need to reinterpret the `cmd_ptr` to the right function
 * pointer type. */

#define CHARS_MAX (sizeof(char) * 256)

struct cmd_trie_node {
    /* The function pointer of the command */
    bool (*cmd_ptr)(struct lkt_state *, size_t, char *[LKT_MESSAGE_ARGS_MAX], int);

    /* Childrens, a byte is 256 possible values */
    struct cmd_trie_node *children[CHARS_MAX];
};

/* The cmd_trie root for all the liblektor, it's OK because we have only one
 * list of possible commands. */
static struct cmd_trie_node *cmd_trie_root = NULL;

PRIVATE_FUNCTION struct cmd_trie_node *
cmd_trie_new(void)
{
    struct cmd_trie_node *ret = LKT_ALLOC_STRUCT(cmd_trie_node);

    for (size_t i = 0; i < CHARS_MAX; ++i) {
        ret->children[i] = NULL;
    }

    /* cmd_ptr == NULL && type == NULL => invalid node, for empty tries. */
    ret->cmd_ptr = NULL;

    return ret;
}

PRIVATE_FUNCTION void
cmd_trie_free(struct cmd_trie_node *root)
{
    for (size_t i = 0; i < CHARS_MAX; ++i) {
        if (root->children[i] != NULL) {
            cmd_trie_free(root->children[i]);
        }
    }

    free(root);
}

PRIVATE_FUNCTION void
cmd_trie_insert(struct cmd_trie_node *root, const char *signed_cmd_name,
                bool (*cmd_ptr)(struct lkt_state *, size_t, char *[LKT_MESSAGE_ARGS_MAX], int))
{
    FAIL_UNLESS(root, "Passing an empty trie root");
    unsigned const char *cmd_name = (unsigned const char *)signed_cmd_name;
    size_t cmd_name_index         = 0;
    size_t current_char           = cmd_name[cmd_name_index];

    while (current_char) {
        /* Create the node if it doesn't exist */
        if (root->children[current_char] == NULL) {
            root->children[current_char] = cmd_trie_new();
        }

        /* curr = curr->next */
        cmd_name_index += 1;
        root         = root->children[current_char];
        current_char = cmd_name[cmd_name_index];
    }

    /* Insert the new command */
    if (root->cmd_ptr == NULL)
        root->cmd_ptr = cmd_ptr;
    else
        LOG_FATAL("The command '%s' was already present in the cmd_trie", signed_cmd_name);
}

PRIVATE_FUNCTION struct cmd_trie_node *
cmd_trie_find(struct cmd_trie_node *root, const char *signed_cmd_name)
{
    FAIL_UNLESS(root, "Passing an empty trie root");
    unsigned const char *cmd_name = (unsigned const char *)signed_cmd_name;
    size_t cmd_name_index         = 0;
    size_t current_char           = cmd_name[cmd_name_index];

    while (current_char) {
        /* No node for the command */
        if (root->children[current_char] == NULL) {
            LOG_WARN("CMD_TRIE", "Failed to find the command: %s", signed_cmd_name);
            return NULL;
        }

        cmd_name_index += 1;
        root         = root->children[current_char];
        current_char = cmd_name[cmd_name_index];
    }

    if (root->cmd_ptr == NULL) {
        LOG_WARN("CMD_TRIE", "Failed to find the command: %s", signed_cmd_name);
        return NULL;
    }

    LOG_INFO("CMD_TRIE", "Found command: %s", signed_cmd_name);
    return root;
}

PRIVATE_FUNCTION void
___cmd_trie_print(struct cmd_trie_node *root, char *old_prefix, const size_t length)
{
    char prefix[length + 2]; /* Old + new_char + \0 */
    memcpy(prefix, old_prefix, sizeof(char) * length);
    prefix[length + 1] = '\0';
    prefix[length]     = '\0';

    /* Found a command? */
    if (root->cmd_ptr != NULL) {
        LKT_OUTPUT("CMD_TRIE", "Got function in trie '%s'", prefix);
    }

    for (size_t i = 0; i < CHARS_MAX; ++i) {
        /* Rec-call with new children */
        if (root->children[i] != NULL) {
            prefix[length] = (char)i;
            ___cmd_trie_print(root->children[i], prefix, length + 1);
        }
    }
}

PRIVATE_FUNCTION void
cmd_trie_print(struct cmd_trie_node *root)
{
    FAIL_UNLESS(root, "Passing an empty trie root");
    ___cmd_trie_print(root, NULL, 0);
}

CONSTRUCTOR_FUNCTION
___cmd_trie_init(void)
{
    LOG_INFO("CMD_TRIE", "Init the command trie");
    cmd_trie_root = cmd_trie_new();

#define mpd_command(str_name, c_func) cmd_trie_insert(cmd_trie_root, (str_name), (c_func));
#include "lektor/internal/commands.def"
#undef mpd_command

    cmd_trie_print(cmd_trie_root);
}

DESTRUCTOR_FUNCTION
___cmd_trie_deinit(void)
{
    LOG_INFO("CMD_TRIE", "Free the command trie");
    cmd_trie_free(cmd_trie_root);
}

/* Client structure */

typedef enum {
    LKT_COMMAND_LIST_OFF,
    LKT_COMMAND_LIST_ON,
    LKT_COMMAND_LIST_ON_OK,
} LKT_COMMAND_LIST_MODE;

struct lkt_client {
    char buffer_in[LKT_MESSAGE_MAX + 1];
    size_t buffer_in_len;

    struct lkt_message *buffer_out[BUFFER_OUT_MAX];
    size_t buffer_out_len;

    struct lkt_command command_list[COMMAND_LIST_MAX];
    size_t command_list_len;
    LKT_COMMAND_LIST_MODE command_list_mode;

    MPD_IDLE_FLAG mpd_idle_watch;
    bool request_close;

    bool authentificated;
    int continuation;
};

static inline bool
lkt_close_client(struct lkt_state *srv, size_t c)
{
    return srv->clients[c - 1].request_close = true;
}

inline int
lkt_get_continuation(struct lkt_state *srv, size_t c)
{
    if (srv == NULL || c > srv->fds_len)
        return 0;
    return srv->clients[c - 1].continuation;
}

inline void
lkt_set_continuation(struct lkt_state *srv, size_t c, int i)
{
    srv->clients[c - 1].continuation = i;
}

inline size_t
lkt_remaining_msg(struct lkt_state *srv, size_t c)
{
    return BUFFER_OUT_MAX - srv->clients[c - 1].buffer_out_len;
}

void
lkt_state_send(struct lkt_state *srv, size_t c, struct lkt_message *msg)
{
    struct lkt_client *cli = &srv->clients[c - 1];
    if (cli->buffer_out_len >= BUFFER_OUT_MAX) {
        LOG_ERROR("NETWORK", "buffer_out is full for client %ld", c);
        return;
    }
    cli->buffer_out[cli->buffer_out_len++] = msg;
    srv->fds[c].events |= POLLOUT;
}

static inline void
send_continue(struct lkt_state *srv, size_t c, int i)
{
    lkt_state_send(srv, c, lkt_message_new_fmt("continue: %d\n", i));
}

static void
send_ok(struct lkt_state *srv, size_t c)
{
    struct lkt_message *ok = (srv->clients[c - 1].command_list_mode == LKT_COMMAND_LIST_ON_OK)
                                 ? lkt_message_new_str("list_OK\n")
                                 : lkt_message_new_str("OK\n");
    lkt_state_send(srv, c, ok);
}

static void
send_ack(struct lkt_state *srv, size_t c, const char *command)
{
    lkt_state_send(srv, c, lkt_message_new_fmt("ACK [1@0] {%s} failed\n", command));
}

static void
send_status(struct lkt_state *srv, size_t c, int status, const char *cmd_name)
{
    if (!status) {
        LOG_DEBUG("COMMAND", "Command '%s'", cmd_name);
        send_ok(srv, c);
    } else {
        LOG_DEBUG("COMMAND", "Command '%s' %s", cmd_name,
                  status == 2 ? "is unknown" : "has failed");
        send_ack(srv, c, cmd_name);
    }
}

static int
command_list_begin(struct lkt_state *srv, size_t c, int list_ok)
{
    struct lkt_client *cli = &srv->clients[c - 1];
    if (cli->command_list_mode != LKT_COMMAND_LIST_OFF)
        return -1;

    if (list_ok)
        cli->command_list_mode = LKT_COMMAND_LIST_ON_OK;
    else
        cli->command_list_mode = LKT_COMMAND_LIST_ON;

    return 0;
}

static int
handle_simple_command(struct lkt_state *srv, size_t c, struct lkt_command cmd)
{
    int err                        = 0;
    int continuation               = 0;
    struct cmd_trie_node *cmd_node = NULL;

    /* Commands that are available only if you are not in IDLE mode */
    if (lkt_client_get_mask(srv, c) == MPD_IDLE_NONE) {
        /* First try the 'fast' search with the trie structure */
        cmd_node = cmd_trie_find(cmd_trie_root, cmd.name);
        if (cmd_node != NULL) {
            err = (int)(!cmd_node->cmd_ptr(srv, c, cmd.args, (int)cmd.cont));
        }

        /* Some commands are not in the structure because they have a unique
         * way to handle arguments. */

        else if (STR_MATCH(cmd.name, "close")) {
            err = !lkt_close_client(srv, c);
        }

        else if (STR_MATCH(cmd.name, "ping")) {
            err = 0;
        }

        else if (STR_MATCH(cmd.name, "listplaylistinfo")) {
            err = !command_find(srv, c, cmd.args, cmd.cont, database_search_playlist_init);
        }

        else if (STR_MATCH(cmd.name, "playlistfind") || STR_MATCH(cmd.name, "playlistsearch")) {
            err = !command_find(srv, c, cmd.args, cmd.cont, database_search_queue_init);
        }

        else if (STR_MATCH(cmd.name, "idle")) {
            LOG_DEBUG("CMD-DEBUG", "Try to do the 'idle' command even if not found in trie");
            err = !command_idle(srv, c, cmd.args, 0);
            goto end_no_send_status;
        }

        else if (STR_MATCH(cmd.name, "search") || STR_MATCH(cmd.name, "find") ||
                 STR_MATCH(cmd.name, "list") || STR_MATCH(cmd.name, "listall") ||
                 STR_MATCH(cmd.name, "listallinfo") || STR_MATCH(cmd.name, "listfiles") ||
                 STR_MATCH(cmd.name, "lsinfo")) {
            err = !command_find(srv, c, cmd.args, cmd.cont, database_search_database_init);
        }

        /* Not found! */
        else {
            err = 2;
        }
    }

    /* commands available only in IDLE mode */
    else {
        if (STR_MATCH(cmd.name, "idle")) {
            err = !command_idle(srv, c, cmd.args, 0);
            goto end_no_send_status;
        }

        else if (STR_MATCH(cmd.name, "noidle")) {
            err = !command_noidle(srv, c, NULL, 0);
        }

        /* Not found! */
        else {
            err = 2;
        }
    }

    continuation = lkt_get_continuation(srv, c);
    if (continuation > 0) {
        LOG_DEBUG("NETWORK", "Client should continue from '%d'", continuation);
        send_continue(srv, c, continuation);
    }
    send_status(srv, c, err, cmd.name);
end_no_send_status:
    return err;
}

static int
command_list_end(struct lkt_state *srv, size_t c)
{
    struct lkt_client *cli = &srv->clients[c - 1];
    int err                = 0;

    if (cli->command_list_mode == LKT_COMMAND_LIST_OFF)
        return -1;

    for (size_t i = 0; i < cli->command_list_len; i++) {
        err |= handle_simple_command(srv, c, cli->command_list[i]);
        lkt_command_free(&cli->command_list[i]);
    }

    cli->command_list_mode = LKT_COMMAND_LIST_OFF;
    cli->command_list_len  = 0;

    if (!err)
        send_ok(srv, c);

    return err;
}

static int
handle_command(struct lkt_state *srv, size_t i, struct lkt_command cmd)
{
    struct lkt_client *cli = &srv->clients[i - 1];
    int err                = 0;

    if (STR_MATCH(cmd.name, "command_list_begin"))
        err = command_list_begin(srv, i, 0);
    else if (STR_MATCH(cmd.name, "command_list_ok_begin"))
        err = command_list_begin(srv, i, 1);
    else if (STR_MATCH(cmd.name, "command_list_end"))
        err = command_list_end(srv, i);
    else if (cli->command_list_mode != LKT_COMMAND_LIST_OFF)
        cli->command_list[cli->command_list_len++] = lkt_command_clone(&cmd);
    else
        err = handle_simple_command(srv, i, cmd);

    LOG_DEBUG("COMMAND", "Command result '%s' -> %d", cmd.name, err);
    return err;
}

static int
handle_incoming_data(struct lkt_state *srv, size_t i)
{
    struct lkt_client *cli = &srv->clients[i - 1];

    for (;;) {
        /* Recieve some data. */
        const ssize_t n = recv(srv->fds[i].fd, cli->buffer_in + cli->buffer_in_len,
                               LKT_MESSAGE_MAX - cli->buffer_in_len, 0);

        if (n < 0) {
            if (errno == EWOULDBLOCK || errno == EAGAIN)
                return 0;
            return -1;
        } else if (n == 0)
            return -1;

        cli->buffer_in_len += (size_t)n; /* NOTE: From here n >= 0, save to cast to size_t */

        /* Handle the data. */
        char *buf = cli->buffer_in;
        for (;;) {
            const size_t line_len = strcspn(buf, "\n");
            if (line_len >= cli->buffer_in_len)
                break;
            buf[line_len] = '\0';

            struct lkt_command cmd = lkt_command_parse(buf);
            lkt_command_log_to_debug(&cmd);
            handle_command(srv, i, cmd);

            buf += line_len + 1;
            cli->buffer_in_len -= (line_len + 1);
        }

        if (cli->buffer_in < buf)
            memmove(cli->buffer_in, buf, cli->buffer_in_len);
        else if (cli->buffer_in_len == LKT_MESSAGE_MAX)
            return -1;
    }

    LOG_UNREACHABLE;
}

static inline void
handle_disconnected_client(struct lkt_state *srv, size_t c)
{
    FAIL_IF(c == 0, "No client '0'");
    struct lkt_client *cli = &srv->clients[c - 1];
    for (size_t i = 0; i < cli->buffer_out_len; i++) {
        lkt_message_free(cli->buffer_out[i]);
        cli->buffer_out[i] = NULL;
    }
    LOG_WARN("NETWORK", "Cleaned %d messages for the client %ld", cli->buffer_out_len, c);
    cli->buffer_out_len = 0;
}

static int
handle_outgoing_data(struct lkt_state *srv, size_t c)
{
    struct lkt_client *const cli = &srv->clients[c - 1];
    struct lkt_message *restrict msg;
    int n;

    for (size_t i = 0; i < cli->buffer_out_len; i++) {
        msg = cli->buffer_out[i];
        n   = lkt_message_send(srv->fds[c].fd, msg);
        if (n <= 0) {
            if (errno == EWOULDBLOCK || errno == EAGAIN) {
                cli->buffer_out_len -= i;
                memmove(cli->buffer_out, cli->buffer_out + i, cli->buffer_out_len);
                return 0;
            }
            if (errno == EPIPE) {
                LOG_WARN("NETWORK", "Client %ld is out, free all its messages", c);
                handle_disconnected_client(srv, i);
            }
            return -1;
        }

        lkt_message_free(msg);
        msg = NULL;
    }

    cli->buffer_out_len = 0;
    srv->fds[c].events &= ~POLLOUT;
    return 0;
}

typedef int (*init_socket_func)(const char *, const char *);

static void
init_permission_socket_unix(const char *file, int perms)
{
    errno = 0;
    if (chmod(file, (__mode_t)perms) < 0) {
        LOG_FATAL("INIT", "Failed to set '%s' to 0%o: %s", file, perms, strerror(errno));
    }
    LOG_INFO("INIT", "Set socket '%s' to perms 0%o", file, perms);
}

static inline int
__set_socket_flags(int fd)
{
    int flags = fcntl(fd, F_GETFL, 0);
    if (flags < 0) {
        LOG_ERROR("INIT", "Failed to get socket flag: %s", strerror(errno));
        goto failure;
    }

    if (fcntl(fd, F_SETFL, flags | O_NONBLOCK) < 0) {
        LOG_ERROR("INIT", "Failed to set non-blocking flag: %s", strerror(errno));
        goto failure;
    }

    return 0;
failure:
    return -1;
}

static int
init_listening_socket_unix(const char *file, const char UNUSED *type /* == "unix" */)
{
    if (!STR_NMATCH(type, "unix", 4))
        LOG_FATAL("Incorrect call to the function, 'port' is not set to 'unix'");

    errno = 0;
    if (unlink(file) < 0) {
        /* Fails silently */
        errno = 0;
        LOG_DEBUG("INIT", "Unlink of socket target file %s failed: %s", file, strerror(errno));
    }

    int fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (fd < 0) {
        LOG_ERROR("INIT", "Failed to create unix socket %s: %s", file, strerror(errno));
        goto failure;
    }

    struct sockaddr_un addr;
    addr.sun_family = AF_UNIX;
    memset(addr.sun_path, 0, sizeof(addr.sun_path));
    strncpy(addr.sun_path, file, sizeof(addr.sun_path) - 1);

    if (__set_socket_flags(fd) < 0) {
        LOG_ERROR("INIT", "Failed to set correct flags to socket %s", file);
        goto failure;
    }

    if (bind(fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_un)) < 0) {
        LOG_ERROR("INIT", "Failed to bind unix socket %s: %s", file, strerror(errno));
        goto failure;
    }

    if (listen(fd, LKT_BACKLOG) < 0) {
        LOG_ERROR("NETWORK", "Failed to listen on %s: %s", file, strerror(errno));
        return -1;
    }

    return fd;
failure:
    close(fd);
    return -1;
}

static int
init_listening_socket_net(const char *host, const char *port)
{
    errno                = 0;
    int fd               = -1;
    static const int yes = 1;
    struct addrinfo *available;
    struct addrinfo hints = { 0 };

    hints.ai_family   = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags    = AI_PASSIVE;
    if (getaddrinfo(host, port, &hints, &available) < 0) {
        LOG_ERROR("INIT", "getaddrinfo failed: %s", strerror(errno));
        return -1;
    }

    for (struct addrinfo *p = available; p; p = p->ai_next) {
        fd = socket(p->ai_family, p->ai_socktype, p->ai_protocol);
        if (fd < 0) {
            LOG_ERROR("INIT", "Failed to create tcp socket: %s", strerror(errno));
            continue;
        }

        if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) < 0) {
            LOG_ERROR("INIT", "Failed to reuse address: %s", strerror(errno));
            goto failure;
        }

        if (__set_socket_flags(fd) < 0) {
            LOG_ERROR("INIT", "Failed to set correct flags to socket");
            goto failure;
        }

        if (bind(fd, p->ai_addr, p->ai_addrlen) < 0) {
            close(fd);
            LOG_ERROR("INIT", "Failed to bind tcp socket: %s", strerror(errno));
            continue;
        }

        break;

    failure:
        close(fd);
        fd = -1;
        break;
    }

    freeaddrinfo(available);

    if (fd < 0) {
        LOG_ERROR("NETWORK", "Failed to bind to %s:%s", host ? host : "0.0.0.0", port);
        return -1;
    }

    if (listen(fd, LKT_BACKLOG) < 0) {
        LOG_ERROR("NETWORK", "Failed to listen: %s", strerror(errno));
        return -1;
    }

    LOG_INFO("NETWORK", "Listening on %s:%s", host ? host : "0.0.0.0", port);
    return fd;
}

static int
accept_all(int listen_fd, struct pollfd *fds, size_t fds_max, size_t *fds_len)
{
    struct sockaddr_in peer_addr;
    socklen_t pa_size = sizeof(peer_addr);
    int fd;
    fds += *fds_len;

    for (int n = 0;; n++) {
        fd = accept(listen_fd, (struct sockaddr *)&peer_addr, &pa_size);
        if (fd < 0) {
            if (errno == EWOULDBLOCK || errno == EAGAIN)
                return n;
            if (errno == ECONNABORTED)
                continue;
            perror("accept");
            return -1;
        }

        if (*fds_len == fds_max) {
            LOG_ERROR("NETWORK", "Maximum number '%ld' of file descriptors reached", fds_max);
            break;
        }

        const int flag = fcntl(fd, F_GETFL, 0);
        if (flag < 0) {
            perror("fcntl");
            break;
        }
        if (fcntl(fd, F_SETFL, flag | O_NONBLOCK) < 0) {
            perror("fcntl");
            break;
        }

        const char *host    = inet_ntoa(peer_addr.sin_addr);
        const uint16_t port = htons(peer_addr.sin_port);
        LOG_INFO("NETWORK", "New connection from %s:%d", host, port);

        fds->fd     = fd;
        fds->events = POLLIN;
        fds++;
        (*fds_len)++;
    }

    shutdown(fd, SHUT_RDWR);
    close(fd);
    return -1;
}

static int
handle_network_events(struct lkt_state *srv)
{
    const int n = poll(srv->fds, srv->fds_len, 100);
    if (n < 0 && errno != EAGAIN && errno != EINTR) {
        perror("poll failed");
        return -1;
    }
    if (n < 0)
        return 0;

    if (srv->fds[0].revents & POLLIN) {
        const int n_in = accept_all(srv->fds[0].fd, srv->fds, srv->fds_max, &srv->fds_len);
        if (n_in < 0)
            return -1;
        // Note from here it's safe to cast n_in to size_t or unsigned
        size_t unsigned_n_in    = (size_t)n_in;
        struct lkt_message *msg = lkt_message_new_str("OK MPD " MPD_VERSION "\n");
        for (; unsigned_n_in; unsigned_n_in--) {
            memset(&srv->clients[srv->fds_len - unsigned_n_in - 1], 0, sizeof(struct lkt_client));
            lkt_state_send(srv, srv->fds_len - unsigned_n_in, lkt_message_clone(msg));
        }
        lkt_message_free(msg);
        msg = NULL;
    }

    for (size_t i = 1; i < srv->fds_len; i++) {
        if (!srv->fds[i].revents)
            continue;
        if (srv->fds[i].revents & (POLLHUP | POLLERR)) {
            close(srv->fds[i].fd);
            srv->fds[i].fd = -1;
            srv->fds[i].revents &= ~(POLLHUP | POLLERR);
            handle_disconnected_client(srv, i);
            LOG_WARN("NETWORK", "Client %ld is out", i);
            continue;
        }
        if (srv->fds[i].revents & POLLIN) {
            if (handle_incoming_data(srv, i) < 0) {
                close(srv->fds[i].fd);
                srv->fds[i].fd = -1;
            }
        }
        if (srv->fds[i].revents & POLLOUT) {
            if (handle_outgoing_data(srv, i) < 0) {
                close(srv->fds[i].fd);
                srv->fds[i].fd = -1;
            }
        }
        if (srv->clients[i - 1].request_close) {
            handle_outgoing_data(srv, i);
            close(srv->fds[i].fd);
            srv->fds[i].fd = -1;
            handle_disconnected_client(srv, i);
        }
    }

    for (size_t i = 1; i < srv->fds_len; i++) {
        if (srv->fds[i].fd > 0)
            continue;

        LOG_WARN("NETWORK", "Connection closed by client %ld", i);
        srv->fds[i]         = srv->fds[srv->fds_len - 1];
        srv->clients[i - 1] = srv->clients[srv->fds_len - 2];
        srv->fds_len--;
        i--;
    }

    return 0;
}

static int
handle_idle_events(struct lkt_state *srv)
{
    struct lkt_client *current_client;
    MPD_IDLE_FLAG common_mask;
    char msg_data[LKT_MESSAGE_MAX];
    bool once = false;

    // No new events
    if (srv->mpd_idle_events == MPD_IDLE_NONE)
        return 0;

    for (size_t i = 1; i < srv->fds_len; ++i) {
        current_client = &srv->clients[i - 1];

        // Client is not idle
        if (current_client->mpd_idle_watch == MPD_IDLE_NONE)
            continue;

        common_mask = current_client->mpd_idle_watch & srv->mpd_idle_events;

        if (MPD_IDLE_NONE == common_mask)
            continue;

        memset(msg_data, 0, LKT_MESSAGE_MAX);
        snprintf(msg_data, LKT_MESSAGE_MAX, "changed: ");
        once |= true;

        if (MPD_IDLE_DATABASE & common_mask)
            strncat(msg_data, "database ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_UPDATE & common_mask)
            strncat(msg_data, "update ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_STORED_PLAYLIST & common_mask)
            strncat(msg_data, "stored_playlist ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_PLAYLIST & common_mask)
            strncat(msg_data, "playlist ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_PLAYER & common_mask)
            strncat(msg_data, "player ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_MIXER & common_mask)
            strncat(msg_data, "mixer ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_OUTPUT & common_mask)
            strncat(msg_data, "output ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_OPTIONS & common_mask)
            strncat(msg_data, "options ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_PARTITION & common_mask)
            strncat(msg_data, "partition ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_STICKER & common_mask)
            strncat(msg_data, "sticker ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_SUBSCRIPTION & common_mask)
            strncat(msg_data, "subscritpion ", LKT_MESSAGE_MAX - 1);
        if (MPD_IDLE_MESSAGE & common_mask)
            strncat(msg_data, "message ", LKT_MESSAGE_MAX - 1);

        size_t data_len        = strlen(msg_data);
        msg_data[data_len - 1] = '\n';
        msg_data[data_len]     = '\0';

        LOG_DEBUG("COMMAND", "Update idle client %ld with mask %x", i, common_mask);
        lkt_state_send(srv, i, lkt_message_new_str(msg_data));
        send_ok(srv, i);
    }

    if (once)
        srv->mpd_idle_events = MPD_IDLE_NONE;

    return 0;
}

int
lkt_client_get_mask(struct lkt_state *srv, size_t c)
{
    return (int)srv->clients[c - 1].mpd_idle_watch;
}

void
lkt_client_add_mask(struct lkt_state *srv, size_t c, MPD_IDLE_FLAG add)
{
    if (pthread_mutex_lock(&srv->lock))
        return;
    srv->clients[c - 1].mpd_idle_watch |= add;
    pthread_mutex_unlock(&srv->lock);
}

void
lkt_client_clear_mask(struct lkt_state *srv, size_t c)
{
    if (pthread_mutex_lock(&srv->lock))
        return;
    srv->clients[c - 1].mpd_idle_watch = MPD_IDLE_NONE;
    pthread_mutex_unlock(&srv->lock);
}

inline bool
lkt_client_auth(struct lkt_state *srv, size_t c, bool set)
{
    return srv->clients[c - 1].authentificated |= set;
}

void
lkt_close_server(struct lkt_state *srv)
{
    for (size_t i = 1; i <= srv->fds_len; ++i) {
        close(srv->fds[i].fd);
        srv->fds[i].fd = -1;
        srv->fds[i].revents &= ~(POLLHUP | POLLERR);
        handle_disconnected_client(srv, i);
    }

    free(srv->fds);
    free(srv->clients);

    MOD_PROC(srv->repo_mod, "free");
    MOD_PROC(srv->window_mod, "free");
    lkt_queue_free(&srv->queue);
    database_close_all();

    LOG_INFO("NET", "lektord will now exit");
}

static inline void
handle_queue_events(struct lkt_state *srv)
{
#if defined(LKT_LOG_QUEUE_EVENTS) && LKT_LOG_QUEUE_EVENTS != 0
#define LOG_EVENT(type)                                                                  \
    if (evt.attr != NULL) {                                                              \
        LOG_DEBUG("EVENT", "Got event " #type " (%d) with attr (int: %d, ptr %0*p)",     \
                  LKT_EVENT_##type, evt.attr, sizeof(size_t) * CHAR_BIT / 4, evt.attr);  \
    } else {                                                                             \
        LOG_DEBUG("EVENT", "Got event " #type " (%d) with NULL attr", LKT_EVENT_##type); \
    }
#else
#define LOG_EVENT(type)
#endif

#define CASE(type, func)   \
    case LKT_EVENT_##type: \
        LOG_EVENT(type)    \
        func;              \
        break;
#define ATTR_IS_STATE(state, func)                      \
    {                                                   \
        if ((__LKT_PLAY_##state) == (size_t)evt.attr) { \
            func;                                       \
        }                                               \
    }

    lkt_event evt;
    char *string = NULL;
    char *arguments_to_pass[LKT_MESSAGE_ARGS_MAX];

redo:
    evt = lkt_queue_handle(&srv->queue);
    // clang-format off
    switch (evt.type) {
    CASE(PLAY_POS, {
        string = LKT_ALLOC_ARRAY(char, BUFFER_MAX);
        safe_snprintf(string, BUFFER_MAX, "%ld", (size_t)evt.attr);
        arguments_to_pass[0] = string;
        arguments_to_pass[1] = NULL;
        command_play(srv, 0, arguments_to_pass, 0);
        free(string);
    })

    CASE(PLAY_NEXT, command_next    (srv, 0, NULL, 0))
    CASE(PLAY_PREV, command_previous(srv, 0, NULL, 0))

    CASE(PLAY_TOGGLE, {
        ATTR_IS_STATE(PAUSE,  database_queue_set_paused(srv->db, 1))
        ATTR_IS_STATE(PLAY,   database_queue_set_paused(srv->db, 0))
        ATTR_IS_STATE(TOGGLE, database_queue_toggle_pause(srv->db))
        ATTR_IS_STATE(STOP,   database_queue_stop(srv->db))
    })

    // Note: Convertion order does matter!
    CASE(PROP_VOL,  database_config_queue(srv->db, "volume",   (int)(size_t)evt.attr))
    CASE(PROP_DUR,  database_config_queue(srv->db, "duration", (int)(size_t)evt.attr))
    CASE(PROP_TIME, database_config_queue(srv->db, "elapsed",  (int)(size_t)evt.attr))

    CASE(SKIP_CURRENT, {
        string = LKT_ALLOC_ARRAY(char, PATH_MAX);
        database_queue_skip_current(srv->db, string);
        MOD_CALL(srv->window_mod, "load", string);
        free(string);
    })

    CASE(TOUCH_KARA,      { database_update_touch(srv->db, (int)(size_t)evt.attr);  })
    CASE(DB_UPDATE_TOTAL, { srv->update_total += (size_t)evt.attr;                  })
    CASE(DB_UPDATING,     { srv->is_updating   = (((uint8_t)(size_t)evt.attr) > 0); })
    CASE(DB_UPDATE_TICK, {
        if (srv->update_current != ((size_t)-1))
            srv->update_current++;
        if (srv->update_current >= srv->update_total)
            (LOG_WARN("EVENT", "Force updating state because tikcs exceded the update count"),
             srv->is_updating    = 0,
             srv->update_total   = 0,
             srv->update_current = 0);
        srv->mpd_idle_events |= MPD_IDLE_DATABASE;
    })

    /* The null event / ignored events, just return / continue */
    CASE(PLAY_FILE, {})
    case LKT_EVENT_NULL:
        return;

    /* Something went REALLY wrong */
    default:
        LOG_FATAL("Got unknown type event: %ld", evt.type);
    }
    // clang-format on
    goto redo;

#undef ATTR_IS_STATE
#undef CASE
}

/*******************
 * Signal handlers *
 *******************/

static int ___lkt_signal_INT  = 0; /* SIGINT  => close the server */
static int ___lkt_signal_QUIT = 0; /* SIGQUIT => close the server */
static int ___lkt_signal_ILL  = 0; /* SIGILL  => NoT veRy HappY   */
static int ___lkt_signal_USR1 = 0; /* SIGUSR1 => not used         */
static int ___lkt_signal_USR2 = 0; /* SIGUSR2 => not used         */

PRIVATE_FUNCTION void
___signal_handler(int sig)
{
#define ___HANDLE(signal)                             \
    case SIG##signal:                                 \
        LOG_INFO("SIGNAL", "Got signal SIG" #signal); \
        ___lkt_signal_##signal = 1;                   \
        break;

    switch (sig) {
        ___HANDLE(INT)
        ___HANDLE(ILL)
        ___HANDLE(QUIT)
        ___HANDLE(USR1)
        ___HANDLE(USR2)
    default: LOG_WARN("SIGNAL", "Not handled signal %d", sig); break;
    }

#undef ___HANDLE
}

PRIVATE_FUNCTION EXIT_FUNCTION
___lkt_handle_INT(struct lkt_state *srv)
{
    LOG_INFO("SIGNAL", "Stopping lektord");

    /* Register default handle for SIGINT, so if you double ctrl+c it exit like
     * it crashed. */
    LOG_INFO("SIGNAL", "Now use default handler for SIGINT and SIGQUIT");
    signal(SIGINT, SIG_DFL);
    signal(SIGQUIT, SIG_DFL);

    lkt_close_server(srv);

    exit(EXIT_SUCCESS);
}

PRIVATE_FUNCTION EXIT_FUNCTION
___lkt_handle_ILL(struct lkt_state UNUSED *srv)
{
    LOG_ERROR("SIGNAL", "Handle the SIGILL, I will try to exit normally");
    ___lkt_handle_INT(srv);
}

static void
___lkt_handle_USR1(struct lkt_state UNUSED *srv)
{
    LOG_INFO("SIGNAL", "Ignore signal SIGUSR1");
}

static void
___lkt_handle_USR2(struct lkt_state UNUSED *srv)
{
    LOG_INFO("SIGNAL", "Ignore signal SIGUSR2");
}

static inline void
handle_signals(struct lkt_state *srv)
{
#define HANDLE(srv, signal)                                                    \
    if (___lkt_signal_##signal) {                                              \
        LOG_INFO("SIGNAL", "Server in handle phase for signal SIG" #signal);   \
        if (srv->___signal_##signal)                                           \
            srv->___signal_##signal(srv);                                      \
        else                                                                   \
            LOG_WARN("SIGNAL", "No signal handler in server for SIG" #signal); \
    }                                                                          \
    ___lkt_signal_##signal = 0;

    HANDLE(srv, INT)
    HANDLE(srv, ILL)
    HANDLE(srv, QUIT)
    HANDLE(srv, USR1)
    HANDLE(srv, USR2)

#undef HANDLE
}

/****************************
 * The server main function *
 ****************************/

EXIT_FUNCTION
lkt_listen(struct lkt_state *srv)
{
    int fds_max_config = 16; /* Default value */
    if (!database_config_get_int(srv->db, "server", "max_clients", &fds_max_config))
        LOG_FATAL("Failed to get the 'server:max_client' field from the configuration");
    if (fds_max_config <= 0)
        LOG_FATAL("The max_clients number must be a positive integer");

    srv->fds_max = (size_t)fds_max_config;
    srv->fds     = LKT_ALLOC_STRUCT_ARRAY(pollfd, srv->fds_max);
    srv->clients = LKT_ALLOC_STRUCT_ARRAY(lkt_client, srv->fds_max);

    LOG_INFO("INIT", "Register signal handlers for signals...");
    srv->___signal_INT  = ___lkt_handle_INT;
    srv->___signal_QUIT = ___lkt_handle_INT;
    srv->___signal_USR1 = ___lkt_handle_USR1;
    srv->___signal_USR2 = ___lkt_handle_USR2;
    srv->___signal_ILL  = ___lkt_handle_ILL;
    signal(SIGINT, ___signal_handler);
    signal(SIGILL, ___signal_handler);
    signal(SIGUSR1, ___signal_handler);
    signal(SIGUSR2, ___signal_handler);
    signal(SIGQUIT, ___signal_handler);
    LOG_INFO("INIT", "Signal handlers registered");

    int is_unix_socket = STR_NMATCH(srv->port, "unix", 4);
    char perms_s[6]    = { 0 };
    int perms          = 0600;

    /* FIXME: Use the server register, and add a socket type (which is
     * unix/tcp) and place the 'socket name' in a [server.socket] and not
     * [server.host]. On the config, the type 'unix/tcp' will result in a
     * function call 'create_unix', 'create_tcp' as [server.use] or something
     * line that. */
    if (is_unix_socket) {
        if (!database_config_get_text(srv->db, "server", "permission", perms_s, 6))
            LOG_WARN("INIT", "Failed to get server socket permission, using default 0%o", perms);
        else {
            perms = (int)strtol(perms_s, NULL, 8); /* In octal */
            LOG_DEBUG("INIT", "Got socket permission 0%o from config file", perms);
        }
    }

    init_socket_func init_sock =
        is_unix_socket ? init_listening_socket_unix : init_listening_socket_net;

    if ((srv->fds[0].fd = init_sock(srv->host, srv->port)) < 0)
        LOG_FATAL("Failed to init socket: %s:%s", srv->host, srv->port);
    if (is_unix_socket)
        init_permission_socket_unix(srv->host, perms);

    srv->fds[0].events = POLLIN;
    srv->fds_len       = 1;
    srv->start_date    = time(NULL);

    if (srv->start_date == (time_t)-1)
        LOG_FATAL("Overflow or error on call to time function");

    /* Listen */
    for (;;) {
        handle_signals(srv);
        if (handle_network_events(srv) < 0) {
            LOG_ERROR("LISTEN", "Failed to handle_network_events");
            break;
        }
        if (handle_idle_events(srv) < 0) {
            LOG_ERROR("LISTEN", "Failed to handle_idle_events");
            break;
        }
        handle_queue_events(srv);
    }

    LOG_FATAL("Infinit event loop break, should not happen");
}
