use super::*;

pub type StatusDeamonMessageType = (
    LektorPlaybackStatusResponse,
    Option<LektorCurrentKaraResponse>,
);

pub struct StatusDeamon {
    thread: Arc<Mutex<Cell<Option<thread::JoinHandle<()>>>>>,
    quit: Arc<AtomicBool>,
    joined: Arc<AtomicBool>,
}

impl Deamon for StatusDeamon {
    type Channels = (Receiver<StatusDeamonMessageType>,);
    implement_deamon_quit!();
    implement_deamon_joined!();

    fn spawn(hostname: String, port: i16) -> io::Result<(Self::Channels, StatusDeamon)> {
        let mut connexion = LektorConnexion::new(hostname, port)?;

        let (responses_send, responses_recv) = channel();
        let quit = Arc::<AtomicBool>::new(AtomicBool::default());
        let joined = Arc::<AtomicBool>::new(AtomicBool::default());
        quit.store(false, atomic::Ordering::SeqCst);
        joined.store(false, atomic::Ordering::SeqCst);
        let quit_deamon = quit.clone();

        let thread = thread::spawn(move || {
            loop {
                break_when_flagged!(quit_deamon);
                thread::sleep(Duration::from_secs(1));
                break_when_flagged!(quit_deamon);

                let status = match connexion.send_query(LektorQuery::PlaybackStatus) {
                    Ok(LektorResponse::PlaybackStatus(res)) => res,
                    Ok(_) => {
                        error!("got invalid response from lektor, not a status...");
                        info!("quiting the status deamon");
                        quit_deamon.store(false, atomic::Ordering::SeqCst);
                        break;
                    }
                    Err(e) => {
                        error!("failed to send the playback status command to lektor: {e}");
                        info!("quiting the status deamon");
                        quit_deamon.store(false, atomic::Ordering::SeqCst);
                        break;
                    }
                };

                let current = match status.state() {
                    LektorState::Stopped => None,
                    LektorState::Play(_) | LektorState::Pause(_) => {
                        match connexion.send_query(LektorQuery::CurrentKara) {
                            Ok(LektorResponse::CurrentKara(res)) => Some(res),
                            Ok(_) => {
                                error!("got invalid response from lektor, not a current kara...");
                                None
                            }
                            Err(e) => {
                                error!("failed to send the current kara command to lektor: {e}");
                                None
                            }
                        }
                    }
                };

                info!("send {status:?} and {current:?}");
                if let Err(e) = responses_send.send((status, current)) {
                    error!("Failed to send a status response to amadeus: {e}");
                    info!("quiting the status deamon");
                    quit_deamon.store(false, atomic::Ordering::SeqCst);
                    break;
                }
            }
            info!("quiting the status deamon!");
        });

        let ret = Self {
            thread: Arc::new(Mutex::new(Cell::new(Some(thread)))),
            quit,
            joined,
        };
        Ok(((responses_recv,), ret))
    }
}
