use crate::{
    constants, utils,
    widgets::{self, render_action_menu},
};
use amadeus_lib::actions;
use lkt_lib::{Kara, LektorType, Playlist};
use std::collections::{HashMap, HashSet};

static DEFAULT_HASHSET_SIZE: usize = 10;

pub struct PlaylistsStore {
    /// The list of all playlists
    playlists: HashMap<String, u64>,

    /// The kara that are contained in the playlists. They are indexed by the
    /// playlist's unique id. The not ordered and unique kara per playlist
    /// aspects of playlists are enforced by the HashSet.
    contents: HashMap<u64, HashSet<Kara>>,

    /// Store the playlists to show.
    playlists_to_show: HashMap<u64, bool>,

    /// The fullfilled actions in this frame.
    actions: Vec<(u64, actions::Action)>,

    /// A temp buffer for handling actions on a specific item, where we know the
    /// id to put but not the render function.
    temp_actions: Vec<actions::Action>,

    /// Whever to refresh all the playlists or not
    refresh_playlists: bool,
}

impl Default for PlaylistsStore {
    fn default() -> Self {
        Self {
            playlists: HashMap::with_capacity(DEFAULT_HASHSET_SIZE),
            contents: HashMap::with_capacity(DEFAULT_HASHSET_SIZE),
            playlists_to_show: HashMap::with_capacity(DEFAULT_HASHSET_SIZE),
            actions: Vec::new(),
            temp_actions: Vec::new(),
            refresh_playlists: false,
        }
    }
}

impl PlaylistsStore {
    pub fn fulfilled_actions(&mut self) -> Vec<(u64, actions::Action)> {
        if !self.actions.is_empty() {
            let ret = self.actions.clone();
            self.actions.clear();
            ret
        } else {
            vec![]
        }
    }

    /// Render the opened playlists.
    pub fn render_playlist_contents(&mut self, ctx: &egui::Context, dark_mode: bool) {
        let to_show = self
            .playlists_to_show
            .iter()
            .filter(|(_, flag)| **flag)
            .map(|(id, _)| *id)
            .collect::<Vec<u64>>();
        for id in to_show {
            let name = self
                .playlists
                .iter()
                .filter(|(_, plt_id)| **plt_id == id)
                .map(|(name, _)| name.as_str())
                .next()
                .unwrap();
            let flag = self.playlists_to_show.get_mut(&id).unwrap();
            let karas = self.contents.get(&id).unwrap();
            let actions = widgets::WindowBuilder::new(name, dark_mode)
                .with_resizable(false)
                .with_default_size(egui::vec2(200., 500.))
                .with_actions(vec![
                    actions::Action::AddPlaylistToQueue,
                    actions::Action::InsertPlaylistToQueue,
                ])
                .with_warning_actions(vec![actions::Action::ClearPlaylistContent])
                .show(flag, ctx, |ui| {
                    egui::ScrollArea::vertical()
                        .hscroll(false)
                        .always_show_scroll(false)
                        .max_width(f32::INFINITY)
                        .show(ui, |ui| {
                            if karas.is_empty() {
                                ui.horizontal(|ui| {
                                    ui.label("Empty playlist");
                                    ui.add_space(constants::PADDING);
                                    ui.add(egui::Spinner::new());
                                });
                            } else {
                                for kara in karas {
                                    ui.horizontal(|ui| {
                                        ui.style_mut().override_text_style =
                                            Some(utils::font::body());
                                        render_action_menu(
                                            ui,
                                            &[
                                                actions::Action::AddKaraToQueue,
                                                actions::Action::InsertKaraInQueue,
                                                actions::Action::AddKaraToPlaylist,
                                                actions::Action::DeleteKaraFromPlaylist(id),
                                                actions::Action::DeleteKaraFromQueue,
                                            ],
                                            widgets::get_label_text_for_kara(kara),
                                            &mut self.temp_actions,
                                        );
                                        self.temp_actions
                                            .iter()
                                            .map(|act| (id, act.clone()))
                                            .for_each(|fullfilled| self.actions.push(fullfilled));
                                        self.temp_actions.clear();
                                    });
                                }
                            }
                        });
                });
            if let Some(actions) = actions {
                actions
                    .iter()
                    .map(|act| (id, act.clone()))
                    .for_each(|fullfilled| self.actions.push(fullfilled));
            }
        }
    }

    /// Render the list of playlists. Also register some of the user's actions
    /// like opening a playlist, etc.
    pub fn render(&mut self, ui: &mut egui::Ui, dark_mode: bool) {
        ui.vertical(|ui| {
            ui.add_space(constants::PADDING);
            ui.horizontal(|ui| {
                ui.style_mut().override_text_style = Some(utils::font::heading1());
                ui.add(egui::widgets::Button::new("🗀 Playlists").frame(false))
                    .clicked()
                    .then(|| self.refresh_playlists = true);
                ui.style_mut().override_text_style = Some(utils::font::small_body());
            });
            ui.style_mut().override_text_style = Some(utils::font::body());
            if self.playlists.is_empty() {
                ui.vertical_centered_justified(|ui| {
                    ui.add(egui::Spinner::new());
                    ui.allocate_space(ui.available_size())
                });
            } else {
                egui::ScrollArea::vertical()
                    .hscroll(false)
                    .always_show_scroll(false)
                    .max_width(f32::INFINITY)
                    .show(ui, |ui| {
                        ui.horizontal(|ui| ui.add_space(ui.available_width()));
                        for (name, id) in &self.playlists {
                            let plt_size = self.playlist_size(name).unwrap();
                            ui.horizontal(|ui| {
                                ui.add_space(constants::PADDING);
                                render_action_menu(
                                    ui,
                                    &[
                                        actions::Action::OpenPlaylist,
                                        actions::Action::AddPlaylistToQueue,
                                        actions::Action::InsertPlaylistToQueue,
                                    ],
                                    "▶",
                                    &mut self.temp_actions,
                                );
                                self.temp_actions
                                    .iter()
                                    .map(|act| (*id, act.clone()))
                                    .for_each(|fullfilled| self.actions.push(fullfilled));
                                self.temp_actions.clear();
                                ui.colored_label(constants::get_text_color(dark_mode), name);
                                ui.style_mut().override_text_style =
                                    Some(utils::font::small_body());
                                ui.label(format!("{plt_size} karas"));
                            });
                        }
                    });
                ui.allocate_space(ui.available_size());
            }
        });
    }

    fn get_playlist_by_name(&self, name: &str) -> Option<(&str, u64)> {
        return self
            .playlists
            .iter()
            .filter(|(plt_name, _)| *plt_name == name)
            .map(|(name, id)| (name.as_str(), *id))
            .next();
    }

    pub fn show_playlist(&mut self, id: u64) {
        if let Some(flag) = self.playlists_to_show.get_mut(&id) {
            *flag = true;
        }
    }

    pub fn clear_playlist(&mut self, id: u64) {
        if let Some(content) = self.contents.get_mut(&id) {
            content.clear();
        }
    }

    pub fn refresh_playlists(&mut self) -> bool {
        let refresh = self.refresh_playlists;
        self.refresh_playlists = false;
        refresh
    }

    pub fn clear_playlists(&mut self) {
        self.contents.clear();
        self.playlists_to_show.clear();
        self.actions.clear();
        self.playlists.clear();
    }

    pub fn create(&mut self, plt: Playlist) {
        self.contents.insert(
            plt.unique_id(),
            HashSet::with_capacity(DEFAULT_HASHSET_SIZE),
        );
        let id = plt.unique_id();
        self.playlists.insert(plt.name, id);
        self.playlists_to_show.insert(id, false);
    }

    pub fn add(&mut self, name: &str, kara: Kara) {
        if let Some((_, id)) = self.get_playlist_by_name(name) {
            if let Some(plt_content) = self.contents.get_mut(&id) {
                plt_content.insert(kara);
            } else {
                panic!("The playlist {name} has no content set, it should have been created when the playlist was created")
            }
        }
    }

    pub fn playlist_size(&self, name: &str) -> Option<usize> {
        let (_, id) = self.get_playlist_by_name(name)?;
        let content = self.contents.get(&id)?;
        Some(content.len())
    }
}
